@@ -1,10 +1,12 @@
 #!/usr/bin/env python
 '''waypoint command handling'''
 
-import time, os, fnmatch
+import time, os, fnmatch, copy, platform
 from pymavlink import mavutil, mavwp
 from MAVProxy.modules.lib import mp_module
-from MAVProxy.modules.lib.mp_menu import *
+from MAVProxy.modules.lib import mp_util
+if mp_util.has_wxpython:
+    from MAVProxy.modules.lib.mp_menu import *
 
 class WPModule(mp_module.MPModule):
     def __init__(self, mpstate):
@@ -16,9 +18,12 @@ class WPModule(mp_module.MPModule):
         self.loading_waypoint_lasttime = time.time()
         self.last_waypoint = 0
         self.wp_period = mavutil.periodic_event(0.5)
+        self.undo_wp = None
+        self.undo_type = None
+        self.undo_wp_idx = -1
         self.add_command('wp', self.cmd_wp,       'waypoint management',
-                         ["<list|clear|move|remove|loop|set>",
-                          "<load|update|save> (FILENAME)"])
+                         ["<list|clear|move|remove|loop|set|undo|movemulti|param>",
+                          "<load|update|save|show> (FILENAME)"])
 
         if self.continue_mode and self.logdir != None:
             waytxt = os.path.join(mpstate.status.logdir, 'way.txt')
@@ -26,9 +31,10 @@ class WPModule(mp_module.MPModule):
                 self.wploader.load(waytxt)
                 print("Loaded waypoints from %s" % waytxt)
 
-        self.menu_added_console = False
-        self.menu_added_map = False
-        self.menu = MPMenuSubMenu('Mission',
+        if mp_util.has_wxpython:
+            self.menu_added_console = False
+            self.menu_added_map = False
+            self.menu = MPMenuSubMenu('Mission',
                                   items=[MPMenuItem('Clear', 'Clear', '# wp clear'),
                                          MPMenuItem('List', 'List', '# wp list'),
                                          MPMenuItem('Load', 'Load', '# wp load ',
@@ -42,6 +48,7 @@ class WPModule(mp_module.MPModule):
                                          MPMenuItem('Draw', 'Draw', '# wp draw ',
                                                     handler=MPMenuCallTextDialog(title='Mission Altitude (m)',
                                                                                  default=100)),
+                                         MPMenuItem('Undo', 'Undo', '# wp undo'),
                                          MPMenuItem('Loop', 'Loop', '# wp loop')])
 
 
@@ -91,8 +98,8 @@ class WPModule(mp_module.MPModule):
         elif mtype in ["WAYPOINT_CURRENT", "MISSION_CURRENT"]:
             if m.seq != self.last_waypoint:
                 self.last_waypoint = m.seq
-                self.say("waypoint %u" % m.seq,priority='message')
-
+                if self.settings.wpupdates:
+                    self.say("waypoint %u" % m.seq,priority='message')
 
 
     def idle_task(self):
@@ -103,10 +110,10 @@ class WPModule(mp_module.MPModule):
                 seq = self.wploader.count()
                 print("re-requesting WP %u" % seq)
                 self.master.waypoint_request_send(seq)
-        if not self.menu_added_console and self.module('console') is not None:
+        if self.module('console') is not None and not self.menu_added_console:
             self.menu_added_console = True
             self.module('console').add_menu(self.menu)
-        if not self.menu_added_map and self.module('map') is not None:
+        if self.module('map') is not None and not self.menu_added_map:
             self.menu_added_map = True
             self.module('map').add_menu(self.menu)
 
@@ -192,6 +199,16 @@ class WPModule(mp_module.MPModule):
             return
         print("Saved %u waypoints to %s" % (self.wploader.count(), filename))
 
+    def get_default_frame(self):
+        '''default frame for waypoints'''
+        if self.settings.terrainalt == 'Auto':
+            if self.get_mav_param('TERRAIN_FOLLOW',0) == 1:
+                return mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT
+            return mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT
+        if self.settings.terrainalt == 'True':
+            return mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT
+        return mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT
+
     def wp_draw_callback(self, points):
         '''callback from drawing waypoints'''
         if len(points) < 3:
@@ -202,8 +219,12 @@ class WPModule(mp_module.MPModule):
         self.wploader.target_system = self.target_system
         self.wploader.target_component = self.target_component
         self.wploader.add(home)
+        if self.get_default_frame() == mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT:
+            use_terrain = True
+        else:
+            use_terrain = False
         for p in points:
-            self.wploader.add_latlonalt(p[0], p[1], self.settings.wpalt, terrain_alt=self.settings.terrainalt)
+            self.wploader.add_latlonalt(p[0], p[1], self.settings.wpalt, terrain_alt=use_terrain)
         self.send_all_waypoints()
 
     def wp_loop(self):
@@ -219,7 +240,6 @@ class WPModule(mp_module.MPModule):
         wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0, 0, mavutil.mavlink.MAV_CMD_DO_JUMP,
                                                           0, 1, 1, -1, 0, 0, 0, 0, 0)
         loader.add(wp)
-        loader.add(loader.wp(1))
         self.loading_waypoints = True
         self.loading_waypoint_lasttime = time.time()
         self.master.waypoint_count_send(self.wploader.count())
@@ -265,6 +285,12 @@ class WPModule(mp_module.MPModule):
             print("No map click position available")
             return
         wp = self.wploader.wp(idx)
+
+        # setup for undo
+        self.undo_wp = copy.copy(wp)
+        self.undo_wp_idx = idx
+        self.undo_type = "move"
+
         (lat, lon) = latlon
         if getattr(self.console, 'ElevationMap', None) is not None and wp.frame != mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT:
             alt1 = self.console.ElevationMap.GetElevation(lat, lon)
@@ -284,6 +310,80 @@ class WPModule(mp_module.MPModule):
         self.wploader.set(wp, idx)
         print("Moved WP %u to %f, %f at %.1fm" % (idx, lat, lon, wp.z))
 
+
+    def cmd_wp_movemulti(self, args):
+        '''handle wp move of multiple waypoints'''
+        if len(args) < 3:
+            print("usage: wp movemulti WPNUM WPSTART WPEND <rotation>")
+            return
+        idx = int(args[0])
+        if idx < 1 or idx > self.wploader.count():
+            print("Invalid wp number %u" % idx)
+            return
+        wpstart = int(args[1])
+        if wpstart < 1 or wpstart > self.wploader.count():
+            print("Invalid wp number %u" % wpstart)
+            return
+        wpend = int(args[2])
+        if wpend < 1 or wpend > self.wploader.count():
+            print("Invalid wp number %u" % wpend)
+            return
+        if idx < wpstart or idx > wpend:
+            print("WPNUM must be between WPSTART and WPEND")
+            return
+
+        # optional rotation about center point
+        if len(args) > 3:
+            rotation = float(args[3])
+        else:
+            rotation = 0
+
+        try:
+            latlon = self.module('map').click_position
+        except Exception:
+            print("No map available")
+            return
+        if latlon is None:
+            print("No map click position available")
+            return
+        wp = self.wploader.wp(idx)
+        if not self.wploader.is_location_command(wp.command):
+            print("WP must be a location command")
+            return
+
+        (lat, lon) = latlon
+        distance = mp_util.gps_distance(wp.x, wp.y, lat, lon)
+        bearing  = mp_util.gps_bearing(wp.x, wp.y, lat, lon)
+
+        for wpnum in range(wpstart, wpend+1):
+            wp = self.wploader.wp(wpnum)
+            if not self.wploader.is_location_command(wp.command):
+                continue
+            (newlat, newlon) = mp_util.gps_newpos(wp.x, wp.y, bearing, distance)
+            if wpnum != idx and rotation != 0:
+                # add in rotation
+                d2 = mp_util.gps_distance(lat, lon, newlat, newlon)
+                b2 = mp_util.gps_bearing(lat, lon, newlat, newlon)
+                (newlat, newlon) = mp_util.gps_newpos(lat, lon, b2+rotation, d2)
+                
+            if getattr(self.console, 'ElevationMap', None) is not None and wp.frame != mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT:
+                alt1 = self.console.ElevationMap.GetElevation(newlat, newlon)
+                alt2 = self.console.ElevationMap.GetElevation(wp.x, wp.y)
+                if alt1 is not None and alt2 is not None:
+                    wp.z += alt1 - alt2
+            wp.x = newlat
+            wp.y = newlon
+            wp.target_system    = self.target_system
+            wp.target_component = self.target_component
+            self.wploader.set(wp, wpnum)
+            
+        self.loading_waypoints = True
+        self.loading_waypoint_lasttime = time.time()
+        self.master.mav.mission_write_partial_list_send(self.target_system,
+                                                        self.target_component,
+                                                        wpstart, wpend+1)
+        print("Moved WPs %u:%u to %f, %f rotation=%.1f" % (wpstart, wpend, lat, lon, rotation))
+
     def cmd_wp_remove(self, args):
         '''handle wp remove'''
         if len(args) != 1:
@@ -294,10 +394,77 @@ class WPModule(mp_module.MPModule):
             print("Invalid wp number %u" % idx)
             return
         wp = self.wploader.wp(idx)
+
+        # setup for undo
+        self.undo_wp = copy.copy(wp)
+        self.undo_wp_idx = idx
+        self.undo_type = "remove"
+
         self.wploader.remove(wp)
         self.send_all_waypoints()
         print("Removed WP %u" % idx)
 
+    def cmd_wp_undo(self):
+        '''handle wp undo'''
+        if self.undo_wp_idx == -1 or self.undo_wp is None:
+            print("No undo information")
+            return
+        wp = self.undo_wp
+        if self.undo_type == 'move':
+            wp.target_system    = self.target_system
+            wp.target_component = self.target_component
+            self.loading_waypoints = True
+            self.loading_waypoint_lasttime = time.time()
+            self.master.mav.mission_write_partial_list_send(self.target_system,
+                                                            self.target_component,
+                                                            self.undo_wp_idx, self.undo_wp_idx)
+            self.wploader.set(wp, self.undo_wp_idx)
+            print("Undid WP move")
+        elif self.undo_type == 'remove':
+            self.wploader.insert(self.undo_wp_idx, wp)
+            self.send_all_waypoints()
+            print("Undid WP remove")
+        else:
+            print("bad undo type")
+        self.undo_wp = None
+        self.undo_wp_idx = -1
+
+    def cmd_wp_param(self, args):
+        '''handle wp parameter change'''
+        if len(args) < 2:
+            print("usage: wp param WPNUM PNUM <VALUE>")
+            return
+        idx = int(args[0])
+        if idx < 1 or idx > self.wploader.count():
+            print("Invalid wp number %u" % idx)
+            return
+        wp = self.wploader.wp(idx)
+        param = [wp.param1, wp.param2, wp.param3, wp.param4]
+        pnum = int(args[1])
+        if pnum < 1 or pnum > 4:
+            print("Invalid param number %u" % pnum)
+            return
+
+        if len(args) == 2:
+            print("Param %u: %f" % (pnum, param[pnum-1]))
+            return
+
+        param[pnum-1] = float(args[2])
+        wp.param1 = param[0]
+        wp.param2 = param[1]
+        wp.param3 = param[2]
+        wp.param4 = param[3]
+        
+        wp.target_system    = self.target_system
+        wp.target_component = self.target_component
+        self.loading_waypoints = True
+        self.loading_waypoint_lasttime = time.time()
+        self.master.mav.mission_write_partial_list_send(self.target_system,
+                                                        self.target_component,
+                                                        idx, idx)
+        self.wploader.set(wp, idx)
+        print("Set param %u for %u to %f" % (pnum, idx, param[pnum-1]))
+
     def cmd_wp(self, args):
         '''waypoint commands'''
         usage = "usage: wp <list|load|update|save|set|clear|loop|remove|move>"
@@ -341,8 +508,14 @@ class WPModule(mp_module.MPModule):
             self.wploader.load(args[1])
         elif args[0] == "move":
             self.cmd_wp_move(args[1:])
+        elif args[0] == "movemulti":
+            self.cmd_wp_movemulti(args[1:])
+        elif args[0] == "param":
+            self.cmd_wp_param(args[1:])
         elif args[0] == "remove":
             self.cmd_wp_remove(args[1:])
+        elif args[0] == "undo":
+            self.cmd_wp_undo()
         elif args[0] == "set":
             if len(args) != 2:
                 print("usage: wp set <wpindex>")
